// This c++ code enumerates set L_a(S) where S is a subset of {0,1}^4/a
// so that kappa(a)\{a} is a subset of S. See the manuscript entitled
// The flip-graph of the 4-dimensional cube is connected (by L. Pournin)
// for comprehesive explanations regarding all the notations, and the
// problem at hand.

#include <vector>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

using namespace std;

// Replaces a string from by a string to in a string str. All the
// occurences of from in str are replaced.
void replaceAll(string& str, const string& from, const string& to)
{
  int i=0;
  while((i = str.find(from, i)) != string::npos)
    {
      str.replace(i, from.length(), to);
      i+=to.length();
    }
}

// Replaces each of the characters in string from by a string to in a string
// str. All the occurences of every characters of from are replaced
void replaceAllChar(string& str, const string& from, const string& to)
{
  for(int j=0;j<from.length();j++)
    replaceAll(str,from.substr(j,1),to);
}

// Checks whether every character of string str2 are found in string str1.
// If yes, return 1. Otherwise returns 0.
int is_subset(string str1,string str2)
{
  int f=0;
  int j;
  for(int i=0;i<str2.length();i++)
    {
      j=0;
      while(j<str1.length() && str1.substr(j,1)!=str2.substr(i,1))
        j++;
      if(j==str1.length())
        return 0;
    }
  return 1;
}

// converts an integer i into a string.
string int2str(int i)
{
  string s;
  stringstream res;
  res << i;
  return res.str();
}

int main()
{
// First, asking which set needs to be enumerated.
// Result is stored in integer variable Si

  int Si;
  cout << "Which set do you want to enumerate?" << "\n";
  cout << "La(S_1) -> 1" << "\n";
  cout << "La(S_2) -> 2" << "\n";
  cout << "La(S_3) -> 3" << "\n";
  cout << "1/2/3 [1]: ";
  cin >> Si;
  if(Si!=1 && Si!=2 && Si!=3)
    {
      Si=1;
    }
  cout << "\n";

// Depending on the value of Si, selects which are the files to be
// loaded (these files are found as outputs of TOPCOM), and how their
// content has to be translated from numbers to characters. 
  string triangs_file,circuits_file,pointconfiguration,vert2char;
  if(Si==1)
    {
      vert2char="bcdefgijklm";
      triangs_file="triangsS1.txt";
      circuits_file="circuitsS1.txt";
      pointconfiguration="{b,c,d,e,f,g,i,j,k,l,m}/a";
    }
  else if(Si==2)
    {
      vert2char="bcefghilmno";
      triangs_file="triangsS2.txt";
      circuits_file="circuitsS2.txt";
      pointconfiguration="{b,c,e,f,g,h,i,l,m,n,o}/a";
    }
  else if(Si==3)
    {
      vert2char="bcdehilmno";
      triangs_file="triangsS3.txt";
      circuits_file="circuitsS3.txt";
      pointconfiguration="{b,c,d,e,h,i,l,m,n,o}/a";
    }
  else
    {
      cout << "We are in BIG trouble there!\n";
      return 1;
    }

// Reads the triangulations file into the vector of strings triangs.
// Each line of this vector contains a string that defines one
// triangulation of set S. For example, the string "bcej ceij"
// represents the triangulation of S that contains the two 
// tetrahedra {b,c,e,j}/a and {c,e,i,j}/a.

  vector<string> triangs;
  cout << "Loading triangulations of point configuration " << pointconfiguration << " from file " << triangs_file << "... ";
  string line;
  ifstream file_handle (triangs_file.c_str());
  if(file_handle.is_open())
    {
      while(file_handle.good())
        {
          getline(file_handle,line);
          replaceAll(line,"},{"," ");
          for(int i=vert2char.length()-1;i>=0;i--)
            replaceAll(line,int2str(i),vert2char.substr(i,1));
          replaceAll(line,",","");
          replaceAll(line,"{{","");
          replaceAll(line,"}}","");
          if(line.length()>2)
            triangs.push_back(line);
        }
      file_handle.close();
    }
  else
    {
      cout << "Unable to open file " << triangs_file << "\n";
      return 1;
    }
  cout << triangs.size() << " triangulations found."<< "\n";


// Reads the file containing the circuits of S into the vector of
// strings tmp. Each line of this vector contains a string that
// defines one circuit of set S and its Radon partition. For
// example, the string "bi j" represents circuit {b,i,j}/a whose
// Radon partition is {{b,i},{j}}/a.

  vector<string> tmp;
  cout << "Loading circuits of point configuration " << pointconfiguration << " from file " << circuits_file << "\n";
  ifstream file_handle2 (circuits_file.c_str());
  if(file_handle2.is_open())
    {
      getline(file_handle2,line); // First line is discarded (from the output format of TOPCOM)
      while(file_handle2.good())
        {
          getline(file_handle2,line);
          replaceAll(line,"},{"," ");
          for(int i=vert2char.length()-1;i>=0;i--)
            replaceAll(line,int2str(i),vert2char.substr(i,1));
          replaceAll(line,",","");
          replaceAll(line,"[{","");
          replaceAll(line,"}]","");
          if(line.length()>2)
            tmp.push_back(line);
        }
      file_handle2.close();
    }
  else
    {
      cout << "Unable to open file " << circuits_file << "\n";
      return 1;
    }

// Reads the file containing the circuits of {0,1}^4 into the vector
// of strings tmp2. Each line in this vector contains a string that
// defines one circuit of {0,1}^4 and its Radon partition in the same
// format used above for the circuits of S. For example, the string
// "bkn ejl" represents circuit {b,e,j,k,l,n} whose Radon partition
// is {{b,k,n},{e,j,l}}.

  vector<string> tmp2;
  string vert4cube="abcdefghijklmnop";
  cout << "Loading circuits of the 4-dimensional cube from file circuits4cube.txt\n";
  ifstream file_handle3 ("circuits4cube.txt");
  if(file_handle3.is_open())
    {
      getline(file_handle3,line); // First line is discarded (from the output format of TOPCOM)
      while(file_handle3.good())
        {
          getline(file_handle3,line);
          replaceAll(line,"},{"," ");
          for(int i=vert4cube.length()-1;i>=0;i--)
            replaceAll(line,int2str(i),vert4cube.substr(i,1));
          replaceAll(line,",","");
          replaceAll(line,"[{","");
          replaceAll(line,"}]","");
          if(line.length()>2)
            tmp2.push_back(line);
        }
      file_handle3.close();
    }
  else
    {
      cout << "Unable to open file circuits4cube.txt\n";
      return 1;
    }

  // Finds the circuits z/a where z belongs to Z(a). These circuits are
  // stored in the vector of strings circuits. Each line in this vector
  // contains a string that defines one such circuit, its Radon partition,
  // epsilon_a^-(z)/a and epsilon_a^+(z)/a (see the manuscript for details).
  // For example, the string "fl dej" represents circuit z/a={d,e,f,j,l}/a,
  // so that epsilon_a^-(z)/a and epsilon_a^+(z)/a are respectively {f,l}/a
  // and {d,e,j}/a.

  vector<string> circuits;
  int k;
  string part1,part2,cir1,cir2;
  cout << "Computing the circuits of point configuration " << pointconfiguration << " that belong to Z(a)/a... ";
  for(int i=0;i<tmp.size();i++)
    {
      k=(tmp[i]).find(" ", 0);
      part1=(tmp[i]).substr(0,k);
      part2=(tmp[i]).substr(k+1,(tmp[i]).length()-k-1);

      cir1="a"+part1+" "+part2;
      cir2="a"+part2+" "+part1;
      for(int j=0;j<tmp2.size();j++)
        {
          if(cir1==tmp2[j])
            circuits.push_back(part1+" "+part2);
          if(cir2==tmp2[j])
            circuits.push_back(part2+" "+part1);
        }
    }

  cout << circuits.size() << " circuits found."<< "\n";

  // Finds all the triangulations in La(S) and store them into the vector
  // of strings La. This is done following  algorithm 1 found in the
  // manuscript. Some variations from algorithm 1 have will be found
  // whose only incidence is on the efficiency of the calculation. They
  // will be mentioned below.

  // Initializes La as a vector of strings.
  vector<string> La;
  string tri,epsilon_minus,epsilon_plus,face,link_face,lambda,lambda2;
  int f1,f2,j,l,l_prev;

  // Index i points to the i-th triangulation of S, found in row i of vector
  // triangs.
  for(int i=0;i<triangs.size();i++)
    {
     // String tri stores the i-th triangulation of S.
     tri=triangs[i];
   
     // Now the circuits in vector circuits will be considered one by one.
     // The flag f1 will be equal to 1 if the circuit considered prevents the
     // triangulation tri to be in La, and to 0 otherwise. This flag is
     // initially set to 0, and the search will finish as soon as a circuit is
     // found that prevents triangulation tri to be in La.
     f1=0;
   
     // Index j points to the j-th circuit, found in row j of vector circuits
     // It is initially equal to 0 and will be incremented in the while loop
     // below.
     j=0;

     // Here, the condition in the while loop will cause an exit either if all
     // the circuits have been tested or if a circuit is found that prevents
     // triangulation tri to be in La (this occurs when at the end of the while
     // loop flag f1 is equal to 1).
     while(j<circuits.size() && f1==0)
       {
         // splits circuit j into epsilon_minus and epsilon_plus.
         k=(circuits[j]).find(" ", 0);
         epsilon_minus=(circuits[j]).substr(0,k);
         epsilon_plus=(circuits[j]).substr(k+1,(circuits[j]).length()-k-1);

         // Initialize lambda to the empty string. This string will store
         // the link of the first face of the triangulation of z that admits
         // epsilon_a^-(z)/a as a subset that is also a face of tri.
         lambda="";

         // f2 is initialized to 1 (see manuscript)
         f2=1;

         // The next loop ranges over the vertices of epsilon_plus. The position
         // of the vertex in string epsilon_plus is given by integer k.
         k=0;
 
         // The loop exists if f2=0 or if all the maximal faces of the
         // triangulation of z/a that admit epsilon_a^-(z)/a as a subset
         // have been treated. In the former case, circuit z/a does not
         // prevent tri to be in La(S). In the latter case, it does.
         while(k<epsilon_plus.length() && f2==1)
           {

             // Initialize lambda2 to the empty string. This string will store
             // the link of the second to last face of the triangulation of z
             // that admits epsilon_a^-(z)/a as a subset that is also a face 
             // of tri.
             lambda2="";

             // builds a maximal face of the triangulation of z that admits
             // epsilon_a^-(z)/a as a subset. This face is obtained removing
             // the vertex of epsilon_plus at position k and by merging the
             // result with epsilon_minus. The face is stored in string face
             face=epsilon_minus+epsilon_plus;
             replaceAll(face,epsilon_plus.substr(k,1),"");

             // Initializes f2 to 0 (see the algorithm in the paper). If face
             // is a face of triangulation tri, then f2 will be reset to 1.
             f2=0;

             // In this while loop, it is checked that string face is indeed a
             // face of triangulation tri. Integer l ranges over every maximal
             // face of triangulation tri. In fact, this integer is the starting
             // position of each face in string tri (so it is a multiple of 5).
             l = 0;

             while(l < tri.length())
               {
                 // checks whether all the vertices of face are also
                 // characters of the face of tri made up of the 4 vertices
                 // from position l to position l+3 in string tri.
                 if(is_subset(tri.substr(l,4),face))
                   {
                     // f2 is set to 1 because string face is a face of triangulation tri
                     f2=1;
                   }
                 // l is incremented. Recall that there is a whitespace between the faces
                 // in string tri! Each face has 4 vertices; + a whitespace so 5 characters
                 // have to be skipped.
                 l=l+5;
               }

             // If string face is a face of triangulation tri, then its link in tri is
             // computed
             if(f2==1)
               {

                 // In this while loop, the link of face in tri is computed. Again,
                 // integer l ranges over every maximal face of triangulation tri.
                 // In fact, this integer is the starting position of each face in
                 // string tri (so it is a multiple of 5).
                 l=0;
                 while(l < tri.length())
                   {
                     // Checks whether all the vertices of face are also
                     // characters of the face of tri made up of the 4 vertices
                     // from position l to position l+3 in string tri.
                     if(is_subset(tri.substr(l,4),face))
                       {
                         // Here, k is 0 when no maximal face of the triangulation of z that
                         // admits epsilon_a^-(z)/a as a subset has been treated yet. In this
                         // case, lambda is initialized to the link of this first face.
                         if(k==0)
                           {
                             // Removes all the chatacters of string face from the
                             // face of tri made up of the 4 vertices from position
                             // l to position l+3 in string tri. Note that each
                             // element of lambda is sorted alphabetically because
                             // of the output format of TOPCOM. Two elements are
                             // separated in string lambda by a whitespace.
                             link_face=tri.substr(l,4);
                             replaceAllChar(link_face, face, "");
                             lambda=lambda+link_face+" ";
                           }
                         // Otherwise, the link of face in tri is stored in string lambda2
                         // with the same formatting as string lambda.
                         else
                           {
                             // Removes all the chatacters of string face from the
                             // face of tri made up of the 4 vertices from position
                             // l to position l+3 in string tri. Note that each
                             // element of lambda is sorted alphabetically because
                             // of the output format of TOPCOM. Two elements are
                             // separated in string lambda by a whitespace.
                             link_face=tri.substr(l,4);
                             replaceAllChar(link_face, face, "");
                             lambda2=lambda2+link_face+" ";
                           }
                       }
                     // l is incremented. Recall that there is a whitespace between the faces
                     // in string tri! Each face has 4 vertices; + a whitespace so 5 characters
                     // have to be skipped.
                     l=l+5;
                   }
               }

             // If at least two faces of the triangulation of z that admits epsilon_a^-(z)/a
             // as a subset have been treated, then lambda2 is compared to lambda.

             if(k>0 && f2==1)
               {
                 // Checks whether the two strings have the same length. If not, the two links
                 // are necessarily distinct. If yes, they necessarily have the same number of
                 // elements.
                 if(lambda.length()==lambda2.length())
                   {
                     // Now, each of the elements of lambda are removed from lambda2.
                     l_prev=0;
                     while((l = lambda.find(" ",l_prev))!=string::npos)
                       {
                         replaceAll(lambda2,lambda.substr(l_prev,l+1-l_prev),"");
                         l_prev=l+1;
                       }
                   }
                 // If lambda2 is now empty then the two links are identical...
                 // If not, f2 is reset to 0.
                 if(lambda2!="")
                   f2=0;
               }
             // Increments k to look at the next face of the triangulation of z that
             // admits epsilon_a^-(z)/a as a subset.

             k=k+1;
           }
         // If f2=0, circuit z/a is not flippable in tri, or epsilon_x^-(z)/a does not
         // belong to tri. Otherwise this circuit prevents tri to belong to La(S).
         if(f2==1)
           f1=1;

         // Increments j to investigate the next circuit...
         j=j+1;
       }
     // If the loop exited with f1=1 then z/a is not in La(S), otherwise it is in La(S)
     if(f1==0)
       {
         replaceAll(tri," ",",");
         cout << i+1 << " over " << triangs.size() << ": triangulation {" << tri << "}/a belongs to La(S).\n";
         La.push_back(tri);
       }
     else
       {
         replaceAll(tri," ",",");
         cout << i+1 << " over " << triangs.size() << ": triangulation {" << tri << "}/a is not in La(S) because of circuit {" << epsilon_minus << "," << epsilon_plus << "}/a.\n";
       }
    }
  // Finally displays La(S) and its size.
  cout << "\nFound the following " << La.size() << " triangulations in La(S) :\n";
  for(int i=0;i<La.size();i++)
    cout << i+1 << ": {" << La[i] << "}/a\n";

  return 0;
}
